﻿//<auto-generated/>
#if (XNA)
using Microsoft.Xna.Framework;
#else
using System;
using FarseerGames.FarseerPhysics.Dynamics;
using FarseerGames.FarseerPhysics.Dynamics.Springs;
using FarseerGames.FarseerPhysics.Mathematics;
#endif

namespace FreeGoo.GameEngine.FarSeer
{
    /// <summary>
    /// Tries to move a body to hit a given position without overshooting. For instance used 
    /// to move items physically correct using the mouse. If you use a LinearSpring, the
    /// body will often start circling the target position instead of hitting it, the
    /// objective of MouseSpring is to hit the target at a relative rest. 
    /// </summary>
    public class MouseSpring : Spring
    {
        protected Vector2 _bodyAttachPoint;
        protected Vector2 _bodyWorldAttachPoint;
        protected Vector2 _targetPoint;
        protected Vector2 _targetVelocity;
        protected Vector2 _thrustDirection;
        protected float _maxAcceleration;
        protected float _distance;
        protected float _targetSpeedMatching;
        protected float _targetPointProjection;
        protected bool _bodyAttachPointOffsetFromCenter;
        
        /// <summary>
        /// Gets or sets the body that the spring applies to.
        /// </summary>
        public Body Body { get; set; }

        /// <summary>
        /// Gets or sets the position on the body where the correcting force should be applied.
        /// </summary>
        public Vector2 BodyAttachPoint
        {
            get { return _bodyAttachPoint; }
            set { _bodyAttachPoint = value; _bodyAttachPointOffsetFromCenter = _bodyAttachPoint.LengthSquared() > 0; }
        }
        /// <summary>
        /// Indicates the position, in world coordinates, where the force will be applied
        /// on the body. This is constantly one time step out of date.
        /// </summary>
        public Vector2 BodyWorldAttachPoint
        {
            get { return _bodyWorldAttachPoint; }
            private set { _bodyWorldAttachPoint = value; }
        }

        /// <summary>
        /// Gets or sets the position the body will be moved to.
        /// </summary>
        public Vector2 TargetPoint
        {
            get { return _targetPoint; }
            set { _targetPoint = value; }
        }

        /// <summary>
        /// Gets or sets the velocity the target is moving at. This is used if
        /// TargetPointProjection or TargetSpeedMatching are >0. Otherwise
        /// it's not used.
        /// </summary>
        public Vector2 TargetVelocity
        {
            get { return _targetVelocity; }
            set { _targetVelocity = value; }
        }

        /// <summary>
        /// Gets or sets the target position (in world coordinates) that the
        /// body should be moved to match.
        /// </summary>
        public float TargetPointProjection
        {
            get { return _targetPointProjection; }
            set { _targetPointProjection = Math.Min(1, Math.Max(0, value)); }
        }

        /// <summary>
        /// Determines if the body should try to match the speed of the target as it
        /// hits the target position. Values between 0 and 1 are valid. Requires
        /// target velocity to be updated.
        /// </summary>
        public float TargetSpeedMatching
        {
            get { return _targetSpeedMatching; }
            set { _targetSpeedMatching = Math.Min(1, Math.Max(0, value)); }
        }

        /// <summary>
        /// Determines how fast the speed of the body should be allowed to change. This value
        /// depends on the size of your world. 
        /// </summary>
        /// Used to internally compute 1) how long it should
        /// take to hit the target and 2) how much force should be applied given the weight of
        /// the body. If the body is attached to other bodies, this acceleration will never be
        /// achieved.
        public float MaxAcceleration
        {
            get { return _maxAcceleration; }
            set { _maxAcceleration = value; }
        }

        /// <summary>
        /// Indicates the direction the thrust was applied in the previous timestep.
        /// </summary>
        public Vector2 ThrustDirection
        {
            get { return _thrustDirection; }
        }

        /// <summary>
        /// Indicates the distance between the bodies position and the target position.
        /// </summary>        
        public float Distance
        {
            get { return _distance; }
        }

        /// <summary>
        /// When the body is closing in on the target, the thrust intensity can be reduced.
        /// ThrustCutoffDistance determines when thrust reduction cuts in.
        /// </summary>        
        public float ThrustCutoffDistance { get; set; }

        public Vector2 _expectedStopPosition = new Vector2(0, 0);

        public MouseSpring(Body body, Vector2 bodyAttachPoint, Vector2 targetPoint, float maxAcceleration)
        {
            Body = body;
            BodyAttachPoint = bodyAttachPoint;
            TargetPoint = targetPoint;
            MaxAcceleration = maxAcceleration;
            TargetVelocity = Vector2.Zero;
            TargetSpeedMatching = 0.5f;
            TargetPointProjection = 0.5f;
            ThrustCutoffDistance = 0;
            _thrustDirection = new Vector2(0, 0);
        }

        public override void Validate()
        {
            if (Body.IsDisposed)
            {
                Dispose();
            }
        }

        public override void Update(float dt)
        {
            base.Update(dt);

            if (IsDisposed)
            {
                return;
            }

            if (_bodyAttachPointOffsetFromCenter)
            {
                Body.GetWorldPosition(ref _bodyAttachPoint, out _bodyWorldAttachPoint);
            }
            else
            {
                _bodyWorldAttachPoint = Body.Position;
            }

            Vector2 distanceVector = TargetPoint - _bodyWorldAttachPoint;
            _distance = distanceVector.Length();
            SpringError = _distance;
            float distanceToStop;

            float thrustX = GetThrustDirection(Body.LinearVelocity.X, _targetVelocity.X, distanceVector.X, out distanceToStop);
            _expectedStopPosition.X = Body.Position.X + distanceToStop;

            float thrustY = GetThrustDirection(Body.LinearVelocity.Y, _targetVelocity.Y, distanceVector.Y, out distanceToStop);
            _expectedStopPosition.Y = Body.Position.Y + distanceToStop;

            _thrustDirection.X = thrustX;
            _thrustDirection.Y = thrustY;

            // Compute the force required to acheave the acceleration
            float forceMagnitude = _maxAcceleration * Body.Mass;
            Vector2 force = new Vector2(_thrustDirection.X * forceMagnitude, _thrustDirection.Y * forceMagnitude);

            // Apply the force to the body, using the center if the attachpoint isn't offset (optimization)
            if (_bodyAttachPointOffsetFromCenter)
            {
                Body.ApplyForceAtLocalPoint(ref force, ref _bodyAttachPoint);
            }
            else
            {
                Body.ApplyForce(ref force);
            }
        }

        /* 
         * This method could be improved by using a two-dimensional method instead of using a one
         * dimensional method twice. For instance, as it stands, a small required change in X
         * and a large required change in Y will both give a thrust of 1 in both direction. The
         * thrust should be scaled so that the dimension that requires the largest change
         * should also receive the largest thrust.
         * 
         */

        public float GetThrustDirection(float speed, float targetSpeed, float distance, out float distanceToStop)
        {
            distanceToStop = 0;

            // Attempt to hit the target going at the same speed as the target. This needs further work,
            // but the objective is to hit the target at a relative rest - doing the same speed.
            // TODO: Make sure this works correctly through some kind of benchmark tests
            speed = speed - targetSpeed * _targetSpeedMatching;

            /* 
             * Compute how long to stop given the current speed. This is used to compute distance
             * and to determine if we're;
             *  - going the wrong way: give max thrust in reverse 
             *  - if we'll stop short: give max thrust 
             *  - if we'll overshoot: give max thrust in reverse
             *  
             */
            float timeToStop = speed / MaxAcceleration;

            // Project where the target will be and try to hit it's future position, not the current 
            // position, this can seriously break down if the target changes direction frequently and
            // quickly because it'll try to hit a position that's no longer relevant. Then it's
            // better to try to hit the current position and see what happens...
            // TODO: Make sure this works correctly through some kind of benchmark tests
            distance = distance + targetSpeed * Math.Abs(timeToStop) * _targetPointProjection;

            float sign = 1;
            if (distance < 0)
            {
                sign = -1;
                speed = -speed;
                distance = -distance;
            }

            if (speed < 0 && distance > 0)
            {
                // Speed and distance are pointing in different directions, just reverse 
                //if (true)//Math.Abs(timeToStop) < 0.2 && distance < ThrustCutoffDistance)
                if (Math.Abs(distance) < ThrustCutoffDistance)
                {
                    // This will not work if the body is moving fast
                    float val = (Math.Abs(distance) / ThrustCutoffDistance);
                    return sign * val * val;
                }
                else
                {
                    return sign;
                }
            }

            timeToStop = Math.Abs(timeToStop);

            distanceToStop = speed * timeToStop - MaxAcceleration * timeToStop * timeToStop / 2;
            float thrust = sign * (distanceToStop >= distance ? -1 : +1);
            distanceToStop = sign * distanceToStop;

            float distanceDifference = Math.Abs(distance - distanceToStop);
            if (Math.Abs(distanceDifference)<ThrustCutoffDistance)
            {
                float val = Math.Abs(distanceDifference) / ThrustCutoffDistance;
                thrust = thrust * val * val;
            }

            return thrust;


            // Here I'd like to add some kind of softening factor so that once the body is close to the
            // target, it can soften it's thrust instead of constantly doing bang-bang corrections.
            /*if (Math.Abs(distanceToStop) < Math.Abs(distance) * 2)
                return bang*0.1f;
            else
                return bang;*/
        }
    }
}
